import json
import itertools 
from mistralai.client import MistralClient
from mistralai.models.chat_completion import ToolCall as MistralToolCall
from typing import List, Generator, Dict, Any
from .connector_interface import Connector
from ..messages import Messages, ToolCall, ChatMessagePiece


class MistralConnector(Connector):
    def __init__(self) -> None:
        self.supports_json_response = True
        self.client = MistralClient()

    def _tool_calls_from_chunks(self, chunks) -> Generator[ChatMessagePiece, None, None]:
        tool_calls = list()
        for chunk in chunks:
            mistral_tool_calls: List[MistralToolCall] = chunk.choices[0].delta.tool_calls
            for t in mistral_tool_calls:
                tool_calls.append(ToolCall(
                    call_id=t.id,
                    function=t.function.name,
                    arguments=json.loads(t.function.arguments)
                ))
        yield ChatMessagePiece(tool_calls=tool_calls)

    def get_streaming_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        response = self.client.chat_stream(
            model=model, 
            temperature=temperature,
            messages=messages.to('mistral'), 
            tools=tools)

        chunk = None
        for chunk in response:
            if chunk.choices[0].delta.content or chunk.choices[0].delta.tool_calls is not None:  # type: ignore
                break
        
        if chunk is None: raise RuntimeError('Got empty completion!')
        chunks = itertools.chain([chunk], response)

        if chunk.choices[0].delta.tool_calls:  # type: ignore
            yield from self._tool_calls_from_chunks(chunks)
        
        else:
            for chunk in chunks:  # type: ignore
                if chunk.choices[0].finish_reason is not None:
                    break

                # in case tools calls begin after some content was sent
                if chunk.choices[0].delta.tool_calls is not None:  
                    remaining_chunks = itertools.chain([chunk], chunks)
                    yield from self._tool_calls_from_chunks(remaining_chunks)

                # got part of message content
                token = chunk.choices[0].delta.content
                if token:
                    yield ChatMessagePiece(content=token)
        
    def get_json_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        if tools:
            yield ChatMessagePiece(warning_message="Mistral models do not support tools when forcing JSON response")
        
        response = self.client.chat(
            model=model, 
            temperature=temperature,
            messages=messages.to('mistral'), 
            tools=tools)
        
        content = response.choices[0].message.content
        if isinstance(content, list):
            content = ' '.join(content)
        yield ChatMessagePiece(content=content)



from anthropic import Anthropic
from anthropic._types import NOT_GIVEN
from typing import List, Generator, Dict, Any
from .connector_interface import Connector
from ..messages import Messages, ChatMessagePiece


class AnthropicConnector(Connector):
    def __init__(self) -> None:
        super().__init__()
        self.client = Anthropic()

    def get_streaming_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        response = self.client.messages.create(
            messages=messages.to('anthropic'),  # type: ignore
            system=messages.system_prompt or NOT_GIVEN,
            temperature=temperature,
            stream=True,
            model=model,
            max_tokens=4096  # maximum defined by Anthropic
        )
        for chunk in response:
            if chunk.type == 'content_block_delta':
                yield ChatMessagePiece(content=chunk.delta.text)
    
    def get_json_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        yield ChatMessagePiece(warning_message="Anthropic models do not support forcing JSON responses")
        yield from self.get_streaming_response(model, temperature, messages, tools)



import json
import itertools 
from openai import OpenAI
from typing import Generator, List, Dict, Any
from .connector_interface import Connector
from ..messages import ToolCall, Messages, ChatMessagePiece


class OpenAIConnector(Connector):
    def __init__(self, api_key: str | None = None, base_url: str | None = None) -> None:
        super().__init__()
        self.client = OpenAI(api_key=api_key, base_url=base_url)

    def _tool_calls_from_chunks(self, chunks) -> Generator[ChatMessagePiece, None, None]:
        tool_calls = list()

        # Build tools_calls from chunks
        for chunk in chunks:
            if chunk.choices[0].finish_reason is not None:
                break
            for i, tool_call in enumerate(chunk.choices[0].delta.tool_calls):
                if len(tool_calls) <= i:
                    tool_call_dict = dict()
                    tool_calls.append(tool_call_dict)
                else:
                    tool_call_dict = tool_calls[i]
                function = chunk.choices[0].delta.tool_calls[0].function
                if function.name is not None:
                    tool_call_dict['id'] = tool_call.id
                    tool_call_dict['function'] = function.name
                    tool_call_dict['arguments'] = ''
                elif function.arguments is not None:
                    tool_call_dict['arguments'] += function.arguments
        
        yield ChatMessagePiece(tool_calls=[ToolCall(call_id=tool['id'], function=tool['function'], arguments=json.loads(tool['arguments'])) for tool in tool_calls])

    def get_streaming_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        completion = self.client.chat.completions.create(
            model=model,
            temperature=temperature,
            messages=messages.to('openai'),  # type: ignore
            tools=tools,                     # type: ignore
            stream=True
        )

        chunk = None
        for chunk in completion:
            if chunk.choices[0].delta.content is not None or chunk.choices[0].delta.tool_calls is not None:  # type: ignore
                break
        
        if chunk is None: raise RuntimeError('Got empty completion!')
        chunks = itertools.chain([chunk], completion)

        if chunk.choices[0].delta.tool_calls:  # type: ignore
            yield from self._tool_calls_from_chunks(chunks)
        
        else:
            for chunk in chunks:  # type: ignore
                if chunk.choices[0].finish_reason is not None:
                    break

                # sometimes gpt calls tools after starting the message
                if chunk.choices[0].delta.tool_calls is not None:  
                    remaining_chunks = itertools.chain([chunk], chunks)
                    yield from self._tool_calls_from_chunks(remaining_chunks)

                # got part of message content
                token = chunk.choices[0].delta.content
                if token:
                    yield ChatMessagePiece(content=token)

    
    def get_json_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        message = self.client.chat.completions.create(
            model=model,
            temperature=temperature,
            messages=messages.to('openai'),  # type: ignore
            tools=tools,                     # type: ignore
            response_format={"type": "json_object"}
        ).choices[0].message

        if message.tool_calls:
            tool_calls: List[ToolCall] = list()
            for tool_call in message.tool_calls:
                tool_calls.append(ToolCall(call_id=tool_call.id, function=tool_call.function.name, arguments=json.loads(tool_call.function.arguments)))
            yield ChatMessagePiece(tool_calls=tool_calls)
        else:
            yield ChatMessagePiece(content=message.content)      



from abc import abstractmethod
from typing import Generator, List, Dict, Any
from ..messages import Messages, ChatMessagePiece


class Connector:
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_streaming_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        raise NotImplementedError()
    
    @abstractmethod
    def get_json_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        raise NotImplementedError()



import sys
from argparse import ArgumentParser
from textwrap import dedent
from . import __version__
from .ui import UI
from .chatbot import Chatbot
from .settings import init_settings
from .framework_models import get_available_frameworks
from .color_logger import get_logger, change_default_log_level
from .utils import path_to_resource


def ascii_art():
    art = dedent(
        """
                                     
        ,------.                              
        |  .---' ,---.,--.  ,--.,---.  ,---.  
        |  `--, | .-. :\  `'  /| .-. :| .-. : 
        |  `---.\   --. \    / \   --.\   --. 
        `------' `----'  `--'   `----' `----'                                

        """)
    print(art)


def _validate_log_level(level: str) -> bool:
    return level.upper() in ['DEBUG', 'INFO', 'WARN', 'WARNING', 'ERROR', 'CRITICAL']


def main():
    parser = ArgumentParser(description="Eevee Chat run-time arguments")
    parser.add_argument('-p', '--port', dest='port', default=4242, type=int, help='Port to run from')
    parser.add_argument('-l', '--log-level', default='INFO', dest='log', help='Set logging level', type=str)
    parser.add_argument('--version', help='Show version', dest='show_version', default=False, action='store_true')
    parser.add_argument('--config-path', help='Show path to config file', dest='config_path', default=False, action='store_true')
    args = parser.parse_args()

    if not _validate_log_level(args.log):
        raise ValueError(f'Invalid log level {args.log}')

    config_path = path_to_resource('config.toml')
    if args.show_version:
        print(f"Eevee Chat version: {__version__}")
    elif args.config_path:
        print(f"Config file path: {config_path}")
    else:
        ascii_art()
        change_default_log_level(args.log.upper())
        get_logger().info(f"Running version: {__version__}")
        init_settings([config_path])
        available_frameworks = get_available_frameworks()
        chatbot = Chatbot(available_frameworks)
        with UI(chatbot, available_frameworks, port=args.port):
            pass


if __name__ == '__main__':
    sys.exit(main())



import os
from typing import List, Dict, Any, Generator
from .openai_connector import OpenAIConnector
from ..messages import Messages, ChatMessagePiece


class DeepSeekConnector(OpenAIConnector):
    def __init__(self) -> None:
        super().__init__(base_url='https://api.deepseek.com/v1', api_key=os.environ['DEEPSEEK_API_KEY'])

    def get_json_response(self, model: str, temperature: float, messages: Messages, tools: List[Dict[str, Any]]) -> Generator[ChatMessagePiece, None, None]:
        yield ChatMessagePiece(warning_message="DeepSeek models do not support forcing JSON responses")
        yield from self.get_streaming_response(model, temperature, messages, tools)



import pathlib

def _get_version_from_setuptools() -> str:
    from pkg_resources import get_distribution

    return get_distribution("eevee-chat").version


__all__ = ["__version__", "ROOT_DIR"]
__version__ = _get_version_from_setuptools()
ROOT_DIR = pathlib.Path(__file__).parent.resolve()
